import numpy as np
import cv2
import calibration as calib
import os
import sys
from tqdm import tqdm

def depth_to_8bit(img, colormap=cv2.COLORMAP_JET, min_thr=None, max_thr=None):
    """
    Auxiliary function to transform depth to a visualizable color map.
    :param img:
    :param colormap:
    :param min_thr:
    :param max_thr:
    :return:
    """
    # threshold extreme values so they do not affect color mapping
    if min_thr:
        img[img < min_thr] = min_thr
    if max_thr:
        img[img > max_thr] = max_thr

    # minmax normalization
    min_val, max_val, _, _ = cv2.minMaxLoc(img, (img > 0).astype(np.uint8))
    img8 = (255 * (img - min_val) / (max_val - min_val)).astype(np.uint8)
    return cv2.applyColorMap(img8, colormap)


def run(capture_path, calib_path, output_path=".", do_rgbdt=True, debug=False, writer=None):
    """
    Conversion of the original capture format of one video sequence to the final multimodal registered frames.
    :param capture_path: Input path of the capturing
    :param calib_path: File containing intrisics of different modalities and extrinsics.
    :param output_path: Output path to calibrated frames
    :param do_rgbdt: If True, computes RGB & thermal overlayed frames to assert quality of registration.
    :param debug: If True, shows visualization in a window of registered frames after saving them to disk.
    :param writer: If cv2.VideoWriter is passed, a video will be created besides the registered frames being saved to disk.
    :return:
    """

    # specification of input captures' structure
    rs_dir = os.path.join(capture_path, "rs")
    pt_dir = os.path.join(capture_path, "pt")
    log_c_path = os.path.join(rs_dir, "color.log")
    log_d_path = os.path.join(rs_dir, "depth.log")
    log_t_path = os.path.join(pt_dir, "thermal.log")

    # read the input capturing frame names, and timetamps
    # timestamps not used here since synchronization is done on-the-fly
    impaths_c, ts_c = calib.read_log(log_c_path, parent=rs_dir)
    impaths_d, ts_d = calib.read_log(log_d_path, parent=rs_dir)
    impaths_t, ts_t = calib.read_log(log_t_path, parent=pt_dir)

    # can happen though: if forced/abrupt stops of recording, there can be some
    # "extra" frames in one of the modalities. cut them out.
    n = np.min([len(impaths_c), len(impaths_d), len(impaths_t)])
    impaths_c, ts_c = impaths_c[:n], ts_c[:n]
    impaths_d, ts_d = impaths_d[:n], ts_d[:n]
    impaths_t, ts_t = impaths_t[:n], ts_t[:n]

    # find realsense device metadata (no metadata for the lepton thermal device)
    rsinfo_path = os.path.join(capture_path, "rs_info.yml")

    # load metadata and calibration parameter files
    rs_info = calib.read_rsinfo(rsinfo_path)
    calib_params = calib.read_calibration(calib_path)

    # read depth scale metadata parameter
    scale_d = rs_info["depth_scale"]

    # read depth and thermal calibration parameters from calibration dict
    calib_d = calib_params["rs/color"]
    calib_t = calib_params["pt/thermal"]
    calib_dims_d = calib_dims_t = calib_params["resize_dims"]   # final registered image dimensions
    ar_d = calib_dims_d[0] / calib_dims_d[1]    # depth aspect ratio
    ar_t = calib_dims_t[0] / calib_dims_t[1]    # thermal aspect ratio

    pbar = tqdm(zip(*[impaths_c, impaths_d, impaths_t]), total=n)

    for impath_c, impath_d, impath_t in pbar:
        # read images
        I_c = cv2.imread(impath_c, cv2.IMREAD_UNCHANGED)
        I_d = cv2.imread(impath_d, cv2.IMREAD_UNCHANGED)
        I_t = cv2.imread(impath_t, cv2.IMREAD_UNCHANGED)

        # transform the aspect ratio (if needed) to match calibration settings
        I_c = calib.pad_to_aspect_ratio(I_c, ar_d)
        I_d = calib.pad_to_aspect_ratio(I_d, ar_d)
        I_t = calib.pad_to_aspect_ratio(I_t, ar_t)

        # correct camera matrices according to read images' dimensions
        K_d = calib.correct_camera_matrix(calib_d["camera_matrix"], calib_dims_d, I_d.shape[::-1])  # numpy (rows,cols) -> cv (x,y)
        K_t = calib.correct_camera_matrix(calib_t["camera_matrix"], calib_dims_t, I_t.shape[::-1])

        # opencv's undistort Python port with modification accepting interpolation
        I_cu = calib.undistort(I_c, K_d, calib_d["dist_coeffs"], interpolation=cv2.INTER_NEAREST)
        I_du = calib.undistort(I_d, K_d, calib_d["dist_coeffs"], interpolation=cv2.INTER_NEAREST)
        I_tu = calib.undistort(I_t, K_t, calib_t["dist_coeffs"], interpolation=cv2.INTER_LINEAR)

        # thermal -> depth alignment. depth's image and camera matrix transformed to thermal dims to make it faster
        I_cu_rs = cv2.resize(I_cu, (I_t.shape[1]*2, I_t.shape[0]*2), interpolation=cv2.INTER_NEAREST)
        I_du_rs = cv2.resize(I_du, (I_t.shape[1], I_t.shape[0]), interpolation=cv2.INTER_NEAREST)
        I_ta = calib.align_to_depth(I_tu, I_du_rs, K_t, calib.correct_camera_matrix(K_d, I_du.shape[::-1], I_tu.shape[::-1]),
                                    scale_d, R_to_d=calib_t["R"], t_to_d=calib_t["t"])

        # save aligned images to disk
        os.makedirs(os.path.join(output_path, 'depth'), exist_ok=True)
        os.makedirs(os.path.join(output_path, 'thermal'), exist_ok=True)
        os.makedirs(os.path.join(output_path, 'thermal-original'), exist_ok=True)

        output_d = impath_d.split('/')[-4] + '_' + impath_d.split('/')[-1]
        output_t = impath_t.split('/')[-4] + '_' + impath_t.split('/')[-1]

        cv2.imwrite(os.path.join(output_path, "depth", output_d), I_du)  # save depth undistorted
        cv2.imwrite(os.path.join(output_path, "thermal", output_t), I_ta)  # save thermal undistorted and aligned
        cv2.imwrite(os.path.join(output_path, "thermal-original", output_t), I_t)

        # visualize aligned thermal in the range of 10-40 degrees celsius, above and below are thresholded
        if do_rgbdt:
            I_ta = depth_to_8bit(I_ta, min_thr=27315, max_thr=30315)
            I_ta = cv2.resize(I_ta, (I_t.shape[1]*2, I_t.shape[0]*2), interpolation=cv2.INTER_NEAREST)
            I_ta_lbl = cv2.addWeighted(I_ta, 0.6, I_cu_rs, 0.6, 0)

            os.makedirs(os.path.join(output_path, "thermal_cmap"), exist_ok=True)
            cv2.imwrite(os.path.join(output_path, "thermal_cmap", output_t), I_ta_lbl)

            if debug:
                cv2.imshow("Aligned thermal", I_ta_lbl)
                cv2.waitKey(1)
            if writer:
                writer.write(I_ta)


if __name__ == "__main__":
    # Input parameters
    # (1) YAML file containing calibration parameters
    # (2) Input capture path
    # (3) Optional: output path (if not specified, current directory)
    #
    # e.g:
    #   python3 align.py extrinsics.new-calibration-blue-small-v0.yml   \
    #                    /Users/aclapes/Data/valid                      \
    #                    /Users/aclapes/Data/valid-out
    #
    calib_path = sys.argv[1]        # YAML file containing parameters
    capture_path = sys.argv[2]      # input (capture) path
    if len(sys.argv) > 3:
        output_path = sys.argv[3]   # output path

    # DBG: create a video
    # fourcc = cv2.VideoWriter_fourcc(*'MP4V')
    # writer = cv2.VideoWriter("video.mp4", fourcc, 10.0, (213, 120))

    # iterate over all captured sequences, each in a separate folder and generate registered multimodal frames
    paths = sorted(os.listdir(capture_path))
    for i, p in enumerate(paths):
        if os.path.isdir(os.path.join(capture_path, p)):
            run(os.path.join(capture_path, p), calib_path, output_path=output_path, do_rgbdt=False, writer=None)

    # writer.release()

    quit()