import numpy as np
import cv2
import time
import warnings

def read_calibration(filepath):
    calib_params = dict()

    fs = cv2.FileStorage(filepath, cv2.FILE_STORAGE_READ)
    if not fs.isOpened():
        return None
    else:
        num_prefixes = int(fs.getNode("num_prefixes").real())
        for i in range(1, num_prefixes+1):
            prefix = fs.getNode(f"prefix-{i}").string()

            calib_params[prefix] = dict(
                camera_matrix=fs.getNode(f"camera_matrix-{i}").mat(),
                dist_coeffs=fs.getNode(f"dist_coeffs-{i}").mat()
            )

            if i > 1:
                calib_params[prefix]["R"] = fs.getNode("R").mat()
                calib_params[prefix]["t"] = fs.getNode("T").mat()

        # for md in fs.getNode("modality-1").string().split(','):
        #   calib_params[md] = dict(
        #     camera_matrix = fs.getNode("camera_matrix-1").mat(),
        #     dist_coeffs = fs.getNode("dist_coeffs-1").mat(),
        #   )
        #
        # for md in fs.getNode("modality-2").string().split(','):
        #   calib_params[md] = dict(
        #     camera_matrix = fs.getNode("camera_matrix-2").mat(),
        #     dist_coeffs = fs.getNode("dist_coeffs-2").mat(),
        #     R = fs.getNode("R").mat(),
        #     t = fs.getNode("T").mat()
        #   )

        calib_params["resize_dims"] = (int(fs.getNode("resize_dims").at(0).real()), int(fs.getNode("resize_dims").at(1).real()))
        calib_params["sync_delay"] = fs.getNode("sync_delay").real()
        calib_params["flags"] = int(fs.getNode("flags").real())
        calib_params["rms"] = fs.getNode("rms").real()

        fs.release()

        return calib_params


def camera_matrix_to_intrinsics(K):
    '''

    fx = K[0,0]
    fy = K[1,1]
    cx = K[0,2]
    cy = K[1,2]

    :param K: camera matrix
    :return: (fx, fy), (cx, cy)
    '''
    return (K[0,0], K[1,1]), (K[0,2], K[1,2])


def parallax(I, map_x, map_y):
    lininds = np.reshape(np.arange(np.prod(I.shape[:2])), (I.shape[0], I.shape[1]))
    I = cv2.remap(I, map_x, map_y, cv2.INTER_NEAREST)
    li_rm = cv2.remap(lininds, map_x, map_y, cv2.INTER_NEAREST, borderMode=cv2.BORDER_CONSTANT, borderValue=-1)
    # li_rm[I_d == 0] = -1P
    li_flat = li_rm.flatten()
    li_order = np.argsort(li_flat)
    d_flat = I_d.flatten()  # flat spatial dimensions
    i_flat = np.reshape(I, (np.prod(I.shape[:2]), I.shape[2]))  # flat spatial dimensions, but not channels
    li_reorder = np.argsort(li_order)
    i_flat_ordered = i_flat[li_order]
    d_flat_ordered = d_flat[li_order]
    uniqs, inds, counts = np.unique(li_flat[li_order], return_counts=True, return_index=True)  # TODO: remove inverse
    K = np.max(counts[1:])
    for k in range(2, K + 1):
        km_mask = counts == k  # k-times mappings
        inds_k = np.repeat(inds[km_mask][:, np.newaxis], k, axis=-1)
        for j in range(k): inds_k[:, j] += j
        P = d_flat_ordered[inds_k]  # parallax depths
        P_min = np.min(P, axis=1)
        del_mask = (P < 50 + P_min[:, np.newaxis]).astype(np.uint8).flatten()  # parallax delete mask
        i_flat_ordered[inds_k.flatten()] = del_mask[:, np.newaxis] * i_flat_ordered[inds_k.flatten()]

    i_flat = i_flat_ordered[li_reorder]
    return np.flipud(np.reshape(i_flat, I.shape))



def get_depth_mappings(I_d, K, Kd, scale_d, R, t):
    '''
    Align some other modality to depth. The slow more-readable version of this code whould be: align_to_depth_slow,
    which is commented underneath.
    :param I_d: depth image
    :param K: other's modality camera matrix
    :param Kd: depth's modality camera matrix
    :param scale_d: a scaling factor to convert depth values into meters
    :param R: other-to-depth rotation matrix
    :param t: other-to-depth translation vector
    :return: map_x and map_y that can be used in OpenCV's cv2.remap(...)
    '''

    # get mapping to depth
    (fx, fy), (cx, cy) = camera_matrix_to_intrinsics(K.astype(np.float32))
    (fx_d, fy_d), (cx_d, cy_d) = camera_matrix_to_intrinsics(Kd.astype(np.float32))

    i = np.repeat(np.linspace(0, I_d.shape[0] - 1, I_d.shape[0], dtype=np.float32), I_d.shape[1])
    j = np.tile(np.linspace(0, I_d.shape[1] - 1, I_d.shape[1], dtype=np.float32), I_d.shape[0])
    d = np.reshape(I_d, [np.prod(I_d.shape), ]).astype(np.float32)

    z = d * scale_d
    x = ((j - cx_d) * z) / fx_d
    y = ((i - cy_d) * z) / fy_d

    P = np.concatenate([x[np.newaxis,:], y[np.newaxis,:], z[np.newaxis,:]], axis=0)

    P_t = np.matmul(R.astype(np.float32), P) + t.astype(np.float32)

    map_x = np.reshape(P_t[0, :] * fx / P_t[2, :] + cx, I_d.shape)
    map_y = np.reshape(P_t[1, :] * fy / P_t[2, :] + cy, I_d.shape)

    return map_x, map_y

def map_to_depth(I, I_d, map_x, map_y):
    lininds = np.reshape(np.arange(np.prod(I.shape[:2])), (I.shape[0], I.shape[1]))
    I = cv2.remap(I, map_x, map_y, cv2.INTER_NEAREST)
    li_rm = cv2.remap(lininds, map_x, map_y, cv2.INTER_NEAREST, borderMode=cv2.BORDER_CONSTANT, borderValue=-1)
    # li_rm[I_d == 0] = -1P
    li_flat = li_rm.flatten()
    li_order = np.argsort(li_flat)
    d_flat = I_d.flatten()  # flat spatial dimensions
    if len(I.shape) == 3:
        i_flat = np.reshape(I, (np.prod(I.shape[:2]), I.shape[2]))  # flat spatial dimensions, but not channels
    elif len(I.shape) == 2:
        i_flat = I.flatten()
    else:
        raise ValueError
    li_reorder = np.argsort(li_order)
    i_flat_ordered = i_flat[li_order]
    d_flat_ordered = d_flat[li_order]
    uniqs, inds, counts = np.unique(li_flat[li_order], return_counts=True, return_index=True)  # TODO: remove inverse
    K = np.max(counts[1:])
    for k in range(2, K + 1):
        km_mask = counts == k  # k-times mappings
        inds_k = np.repeat(inds[km_mask][:, np.newaxis], k, axis=-1)
        for j in range(k): inds_k[:, j] += j
        P = d_flat_ordered[inds_k]  # parallax depths
        P_min = np.min(P, axis=1)
        del_mask = (P < 50 + P_min[:, np.newaxis]).astype(np.uint8).flatten()  # parallax delete mask
        if len(i_flat_ordered.shape) > 1:
            del_mask = del_mask[:, np.newaxis]
        i_flat_ordered[inds_k.flatten()] = del_mask * i_flat_ordered[inds_k.flatten()]

    i_flat = i_flat_ordered[li_reorder]
    return np.reshape(i_flat, I.shape)


def undistort(src, K, D, interpolation=cv2.INTER_LINEAR):
    dst = np.zeros(src.shape, src.dtype)
    stripe_size0 = int( min(max(1, (1 << 12) / max(src.shape[1], 1)), src.shape[0]) )
    map1 = np.zeros((stripe_size0, src.shape[1], 2), np.int16)
    map2 = np.zeros((stripe_size0, src.shape[1]), np.uint16)
    I = np.eye(3,3)
    A = K.astype(np.double)
    Ar = A.copy()
    dist_coeffs = D.astype(np.double)

    v0 = Ar[1,2]
    for y in range(0, src.shape[0], stripe_size0):
        stripe_size = min(stripe_size0, src.shape[0] - y)
        Ar[1, 2] = v0 - y
        map1_part = map1[0:stripe_size, :] # map1.rowRange(0, stripe_size)
        map2_part = map2[0:stripe_size, :]
        # dst_part = dst[y:y+stripe_size,:]

        map1[0:stripe_size], map2[0:stripe_size,] = cv2.initUndistortRectifyMap(A, dist_coeffs, I, Ar,       \
                                                                                        (src.shape[1], stripe_size), \
                                                                                        cv2.CV_16SC2)
        dst[y:y+stripe_size] = cv2.remap(src, map1_part, map2_part, interpolation, borderMode=cv2.BORDER_CONSTANT)

    return dst


def align_to_depth(I, I_d, K, K_d, scale_d, R_to_d=None, t_to_d=None):
    '''
    Align some other modality to depth.

    :param I: image to align to depth image
    :param I_d: depth image
    :param K: other's modality camera matrix
    :param Kd: depth's modality camera matrix
    :param scale_d: a scaling factor to convert depth values into meters
    :param R: other-to-depth rotation matrix
    :param t: other-to-depth translation vector

    :return: aligned image to depth
    '''

    try:
        if not (R_to_d is None or t_to_d is None):
            map_x, map_y = get_depth_mappings(I_d, K, K_d, scale_d, R_to_d, t_to_d)
            I_mapped = map_to_depth(I, I_d, map_x, map_y)
    except cv2.error:
        warnings.warn("calibration.apply exception")
        return None

    return I_mapped


def pad_to_aspect_ratio(I, ar):
    """
    Transforms I image aspect ratio to match dims.

    :param I: input image to be transformed
    :param ar: target aspect ratio

    :return: image with aspect ratio matching "ar"
    """
    if ar > I.shape[1] / I.shape[0]:
        dims_new = (round(ar * I.shape[0]), I.shape[0])
    elif ar < I.shape[1] / I.shape[0]:
        dims_new = (I.shape[1], round(ar * I.shape[1]))
    else:
        return I

    I = cv2.resize(I, (int(I.shape[1] * (dims_new[1] / I.shape[0])), dims_new[1]))
    padding = int((dims_new[0] - I.shape[1]) / 2)
    I = cv2.copyMakeBorder(I, 0, 0, padding, dims_new[0] - I.shape[1] - padding, 0)

    return I


def correct_camera_matrix(K, dims_k, dims_new):
    K_new = K.copy()
    K_new[0, :] /= (dims_k[0] / dims_new[0])
    K_new[1, :] /= (dims_k[1] / dims_new[1])

    return K_new


def read_rsinfo(filepath):
    info_dict = {}

    fs = cv2.FileStorage(filepath, cv2.FILE_STORAGE_READ)
    info_dict["serial_number"] = fs.getNode("serial_number").string()
    info_dict["depth_scale"] = fs.getNode("depth_scale").real()
    fs.release()

    return info_dict


def read_log(file, parent=""):
    import os

    images = []
    ts = []
    with open(file, 'r') as f:
        for l in f:
            line_split = l.strip('\n').split(',')
            assert(len(line_split) == 2)
            images.append(os.path.join(parent, line_split[0]) if parent != "" else line_split[0])
            ts.append(float(line_split[1]))

    # fallback
    if len(images) == 0:
        p = os.path.splitext(file)[0]
        images = [os.path.join(p, img) for img in sorted(os.listdir(p))]

    return images, ts


if __name__ == "__main__":

    # Some test code

    fs = cv2.FileStorage("data/2019-05-07_13.43.07/extrinsics-parameters.calibration-blue-small.yml", flags=cv2.FILE_STORAGE_READ)
    K_d = fs.getNode("camera_matrix_1").mat()
    K_o = fs.getNode("camera_matrix_2").mat()
    distcoeffs_d = fs.getNode("dist_coeffs_1").mat()
    distcoeffs_o = fs.getNode("dist_coeffs_2").mat()
    R_to_d = fs.getNode("R").mat()
    T_to_d = fs.getNode("T").mat()
    fs.release()

    fs = cv2.FileStorage("data/2019-05-07_13.43.07/rs_info.yml", flags=cv2.FILE_STORAGE_READ)
    scale_d = fs.getNode("depth_scale").real()
    fs.release()

    I_d_p = cv2.imread("data/2019-05-07_13.43.07/rs/depth/d_00019391.png", cv2.IMREAD_UNCHANGED)
    I_t = cv2.imread("data/2019-05-07_13.43.07/pt/thermal/t_00019391.png", cv2.IMREAD_UNCHANGED)

    I_d = cv2.normalize(I_d_p, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8UC1)
    I_d = cv2.applyColorMap(I_d, cv2.COLORMAP_BONE)

    I_t = cv2.normalize(I_t, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8UC1)
    I_t = cv2.resize(I_t, (960, 720))
    I_t = cv2.applyColorMap(I_t, cv2.COLORMAP_HOT)
    padding = int((1280-960)/2)
    I_t = cv2.copyMakeBorder(I_t, 0, 0, padding, padding, 0)

    I_d_p = cv2.undistort(I_d_p, K_d, distcoeffs_d)
    I_t = cv2.undistort(I_t, K_o, distcoeffs_o)

    st_time = time.time()
    map_x_fast, map_y_fast = align_to_depth(I_d_p, K_d, K_o, scale_d, R_to_d, T_to_d)
    print(f"align_to_depth fast took {time.time() - st_time} secs.")

    # st_time = time.time()
    # map_x_slow, map_y_slow = align_to_depth_slow(I_d_p, K_d, K_o, scale_d, R_to_d, T_to_d)
    # print(f"align_to_depth slow took {time.time() - st_time} secs.")

    # assert np.allclose(map_x_slow, map_x_fast) and np.allclose(map_y_slow, map_y_fast)
    map_x = map_x_fast
    map_y = map_y_fast

    st_time = time.time()
    I_t = cv2.remap(I_t, map_x, map_y, cv2.INTER_LINEAR)
    print(f"cv2.remap took {time.time() - st_time} secs.")

    cv2.imshow("window", I_t)
    cv2.waitKey()